/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/evalf/evalf.c                                          *
 * Created:       2007-01-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2007-2008 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: evalf.c 498 2008-08-24 22:40:05Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <math.h>

#include "evalf.h"


static int evalf_literal (const char **str, double *val);
static int evalf_expr (const char **str, double *val);


static const char *argv0 = NULL;


static
void prt_help (void)
{
	fputs (
		"evalf: evaluate floating point expressions\n"
		"\n"
		"usage: evalf [options] [expression...]\n"
		"  -e, --expression expr  Evaluate an expression\n",
		stdout
	);

	fflush (stdout);
}

static
uint64_t dbl_to_uint64 (double v)
{
	uint64_t r;

	if (v < 0.0) {
		r = (uint64_t) (-v);
		r = -r;
	}
	else {
		r = (uint64_t) v;
	}

	return (r);
}

static
int dbl_bit_and (double *v, double x, double y)
{
	*v = dbl_to_uint64 (x) & dbl_to_uint64 (y);

	return (0);
}

static
int dbl_bit_or (double *v, double x, double y)
{
	*v = dbl_to_uint64 (x) | dbl_to_uint64 (y);

	return (0);
}

static
int dbl_bit_xor (double *v, double x, double y)
{
	*v = dbl_to_uint64 (x) ^ dbl_to_uint64 (y);

	return (0);
}

static
int is_digit (char c)
{
	if ((c >= '0') && (c <= '9')) {
		return (1);
	}

	return (0);
}

#if 0
static
int is_alpha (char c)
{
	if ((c >= 'a') && (c <= 'z')) {
		return (1);
	}

	if ((c >= 'A') && (c <= 'Z')) {
		return (1);
	}

	return (0);
}
#endif

static
int match_space (const char **str)
{
	int        r;
	const char *tmp;

	tmp = *str;

	r = 0;
	while ((*tmp == ' ') || (*tmp == '\t')) {
		r = 1;
		tmp += 1;
	}

	*str = tmp;

	return (r);
}

static
int match_string (const char **str, const char *val)
{
	const char *tmp;

	tmp = *str;

	match_space (&tmp);

	while (*val != 0) {
		if (*tmp != *val) {
			return (0);
		}

		tmp += 1;
		val += 1;
	}

	*str = tmp;

	return (1);
}

static
int match_peek_string (const char **str, const char *val)
{
	const char *tmp;

	tmp = *str;

	match_space (&tmp);

	while (*val != 0) {
		if (*tmp != *val) {
			return (0);
		}

		tmp += 1;
		val += 1;
	}

	return (1);
}

#if 0
static
int match_ident (const char **str, char *val, unsigned max)
{
	unsigned   i;
	const char *tmp;

	tmp = *str;

	match_space (&tmp);

	if ((is_alpha (*tmp) == 0) && (*tmp != '_')) {
		return (0);
	}

	if (max < 2) {
		return (0);
	}

	i = 0;
	val[i++] = *(tmp++);

	while (is_alpha (*tmp) || is_digit (*tmp) || (*tmp == '_')) {
		if (i >= max) {
			return (0);
		}

		val[i++] = *(tmp++);
	}

	if (i >= max) {
		return (0);
	}

	val[i] = 0;

	*str = tmp;

	return (1);
}
#endif

static
int match_double (const char **str, double *val)
{
	const char *tmp;
	double     v;

	tmp = *str;

	match_space (&tmp);

	if (is_digit (*tmp) == 0) {
		return (0);
	}

	v = 0.0;
	while (is_digit (*tmp)) {
		v = 10.0 * v + (*tmp - '0');
		tmp += 1;
	}

	if (*tmp == '.') {
		double f, d;

		tmp += 1;

		f = 0.0;
		d = 1.0;

		while (is_digit (*tmp)) {
			f = 10.0 * f + (*tmp - '0');
			d = 10.0 * d;
			tmp += 1;
		}

		v = v + (f / d);
	}

	*val = v;
	*str = tmp;

	return (1);
}

static
int evalf_function_0 (const char **str, double *val, const char *name)
{
	if (strcmp (name, "pi") == 0) {
		*val = 2.0 * asin (1.0);
	}
	else {
		return (1);
	}

	return (0);
}

static
int evalf_function_1 (const char **str, double *val, const char *name)
{
	double x;

	if (evalf_literal (str, &x)) {
		return (1);
	}

	if (strcmp (name, "abs") == 0) {
		*val = fabs (x);
	}
	else if (strcmp (name, "acos") == 0) {
		*val = acos (x);
	}
	else if (strcmp (name, "asin") == 0) {
		*val = asin (x);
	}
	else if (strcmp (name, "atan") == 0) {
		*val = atan (x);
	}
	else if (strcmp (name, "ceil") == 0) {
		*val = ceil (x);
	}
	else if (strcmp (name, "cosh") == 0) {
		*val = cosh (x);
	}
	else if (strcmp (name, "cos") == 0) {
		*val = cos (x);
	}
	else if (strcmp (name, "cot") == 0) {
		*val = cos (x) / sin (x);
	}
	else if (strcmp (name, "exp") == 0) {
		*val = exp (x);
	}
	else if (strcmp (name, "floor") == 0) {
		*val = floor (x);
	}
	else if ((strcmp (name, "log") == 0) || (strcmp (name, "ln") == 0)) {
		if (x <= 0.0) {
			return (1);
		}

		*val = log (x);
	}
	else if (strcmp (name, "log2") == 0) {
		if (x <= 0.0) {
			return (1);
		}

		*val = log (x) / log (2.0);
	}
	else if (strcmp (name, "log10") == 0) {
		if (x <= 0.0) {
			return (1);
		}

		*val = log10 (x);
	}
	else if (strcmp (name, "sinh") == 0) {
		*val = sinh (x);
	}
	else if (strcmp (name, "sin") == 0) {
		*val = sin (x);
	}
	else if (strcmp (name, "sqrt") == 0) {
		if (x < 0.0) {
			return (1);
		}
		*val = sqrt (x);
	}
	else if (strcmp (name, "sqr") == 0) {
		*val = x * x;
	}
	else if (strcmp (name, "tanh") == 0) {
		*val = tanh (x);
	}
	else if (strcmp (name, "tan") == 0) {
		*val = tan (x);
	}
	else {
		return (1);
	}

	return (0);
}

static
int evalf_function_2 (const char **str, double *val, const char *name)
{
	double x, y;

	if (match_string (str, "(") == 0) {
		return (1);
	}

	if (evalf_expr (str, &x)) {
		return (1);
	}

	if (match_string (str, ",") == 0) {
		return (1);
	}

	if (evalf_expr (str, &y)) {
		return (1);
	}

	if (match_string (str, ")") == 0) {
		return (1);
	}

	if (strcmp (name, "atan2") == 0) {
		*val = atan2 (x, y);
	}
	else {
		return (1);
	}

	return (0);
}

static
int evalf_literal (const char **str, double *val)
{
	if (match_double (str, val)) {
		return (0);
	}

	if (match_string (str, "(")) {
		if (evalf_expr (str, val)) {
			return (1);
		}

		if (match_string (str, ")") == 0) {
			return (1);
		}

		return (0);
	}

	if (match_string (str, "abs")) {
		return (evalf_function_1 (str, val, "abs"));
	}
	else if (match_string (str, "acos")) {
		return (evalf_function_1 (str, val, "acos"));
	}
	else if (match_string (str, "asin")) {
		return (evalf_function_1 (str, val, "asin"));
	}
	else if (match_string (str, "atan2")) {
		return (evalf_function_2 (str, val, "atan2"));
	}
	else if (match_string (str, "atan")) {
		return (evalf_function_1 (str, val, "atan"));
	}
	else if (match_string (str, "ceil")) {
		return (evalf_function_1 (str, val, "ceil"));
	}
	else if (match_string (str, "cosh")) {
		return (evalf_function_1 (str, val, "cosh"));
	}
	else if (match_string (str, "cos")) {
		return (evalf_function_1 (str, val, "cos"));
	}
	else if (match_string (str, "cot")) {
		return (evalf_function_1 (str, val, "cot"));
	}
	else if (match_string (str, "exp")) {
		return (evalf_function_1 (str, val, "exp"));
	}
	else if (match_string (str, "floor")) {
		return (evalf_function_1 (str, val, "floor"));
	}
	else if (match_string (str, "ln")) {
		return (evalf_function_1 (str, val, "log"));
	}
	else if (match_string (str, "log10")) {
		return (evalf_function_1 (str, val, "log10"));
	}
	else if (match_string (str, "log2")) {
		return (evalf_function_1 (str, val, "log2"));
	}
	else if (match_string (str, "log")) {
		return (evalf_function_1 (str, val, "log"));
	}
	else if (match_string (str, "pi")) {
		return (evalf_function_0 (str, val, "pi"));
	}
	else if (match_string (str, "sinh")) {
		return (evalf_function_1 (str, val, "sinh"));
	}
	else if (match_string (str, "sin")) {
		return (evalf_function_1 (str, val, "sin"));
	}
	else if (match_string (str, "sqrt")) {
		return (evalf_function_1 (str, val, "sqrt"));
	}
	else if (match_string (str, "sqr")) {
		return (evalf_function_1 (str, val, "sqr"));
	}
	else if (match_string (str, "tanh")) {
		return (evalf_function_1 (str, val, "tanh"));
	}
	else if (match_string (str, "tan")) {
		return (evalf_function_1 (str, val, "tan"));
	}

	return (1);
}

static
int evalf_exp (const char **str, double *val)
{
	double v;

	if (evalf_literal (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "**")) {
			if (evalf_literal (str, &v)) {
				return (1);
			}
			*val = pow (*val, v);
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_neg (const char **str, double *val)
{
	if (match_string (str, "+")) {
		if (evalf_neg (str, val)) {
			return (1);
		}
		return (0);
	}
	else if (match_string (str, "-")) {
		if (evalf_neg (str, val)) {
			return (1);
		}
		*val = -*val;
		return (0);
	}
	else if (match_string (str, "!") || match_string (str, "~")) {
		if (evalf_neg (str, val)) {
			return (1);
		}
		*val = (*val < 0.0) ? 1.0 : -1.0;
		return (0);
	}

	if (evalf_exp (str, val)) {
		return (1);
	}

	return (0);
}

static
int evalf_mul (const char **str, double *val)
{
	double v;

	if (evalf_neg (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "*")) {
			if (evalf_neg (str, &v)) {
				return (1);
			}
			*val = *val * v;
		}
		else if (match_string (str, "/")) {
			if (evalf_neg (str, &v)) {
				return (1);
			}
			if (v == 0.0) {
				return (1);
			}
			*val = *val / v;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_add (const char **str, double *val)
{
	double v;

	if (evalf_mul (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "+")) {
			if (evalf_mul (str, &v)) {
				return (1);
			}
			*val = *val + v;
		}
		else if (match_string (str, "-")) {
			if (evalf_mul (str, &v)) {
				return (1);
			}
			*val = *val - v;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_shift (const char **str, double *val)
{
	double v;

	if (evalf_add (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "<<")) {
			if (evalf_add (str, &v)) {
				return (1);
			}
			*val = *val * pow (2.0, v);
		}
		else if (match_string (str, ">>")) {
			if (evalf_add (str, &v)) {
				return (1);
			}
			*val = *val * pow (2.0, -v);
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_cmp (const char **str, double *val)
{
	double v;

	if (evalf_shift (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "<=")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val <= v) ? 1.0 : -1.0;
		}
		else if (match_string (str, ">=")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val >= v) ? 1.0 : -1.0;
		}
		else if (match_peek_string (str, "<<")) {
			break;
		}
		else if (match_peek_string (str, ">>")) {
			break;
		}
		else if (match_string (str, "<")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val < v) ? 1.0 : -1.0;
		}
		else if (match_string (str, ">")) {
			if (evalf_shift (str, &v)) {
				return (1);
			}
			*val = (*val > v) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_equ (const char **str, double *val)
{
	double v;

	if (evalf_cmp (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "==")) {
			if (evalf_cmp (str, &v)) {
				return (1);
			}
			*val = (*val == v) ? 1.0 : -1.0;
		}
		else if (match_string (str, "!=")) {
			if (evalf_cmp (str, &v)) {
				return (1);
			}
			*val = (*val != v) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_bit_and (const char **str, double *val)
{
	double v;

	if (evalf_equ (str, val)) {
		return (1);
	}

	while (1) {
		if (match_peek_string (str, "&&")) {
			break;
		}

		if (match_string (str, "&")) {
			if (evalf_equ (str, &v)) {
				return (1);
			}
			if (dbl_bit_and (val, *val, v)) {
				return (1);
			}
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_bit_xor (const char **str, double *val)
{
	double v;

	if (evalf_bit_and (str, val)) {
		return (1);
	}

	while (1) {
		if (match_peek_string (str, "^^")) {
			break;
		}

		if (match_string (str, "^")) {
			if (evalf_bit_and (str, &v)) {
				return (1);
			}
			if (dbl_bit_xor (val, *val, v)) {
				return (1);
			}
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_bit_or (const char **str, double *val)
{
	double v;

	if (evalf_bit_xor (str, val)) {
		return (1);
	}

	while (1) {
		if (match_peek_string (str, "||")) {
			break;
		}

		if (match_string (str, "|")) {
			if (evalf_bit_xor (str, &v)) {
				return (1);
			}
			if (dbl_bit_or (val, *val, v)) {
				return (1);
			}
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_and (const char **str, double *val)
{
	double v;

	if (evalf_bit_or (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "&&")) {
			if (evalf_bit_or (str, &v)) {
				return (1);
			}
			*val = ((*val > 0.0) && (v > 0.0)) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_xor (const char **str, double *val)
{
	double v;

	if (evalf_and (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "^^")) {
			if (evalf_and (str, &v)) {
				return (1);
			}
			*val = ((*val > 0.0) != (v > 0.0)) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_or (const char **str, double *val)
{
	double v;

	if (evalf_xor (str, val)) {
		return (1);
	}

	while (1) {
		if (match_string (str, "||")) {
			if (evalf_xor (str, &v)) {
				return (1);
			}
			*val = ((*val > 0.0) || (v > 0.0)) ? 1.0 : -1.0;
		}
		else {
			break;
		}
	}

	return (0);
}

static
int evalf_if (const char **str, double *val)
{
	double v1, v2;

	if (evalf_or (str, val)) {
		return (1);
	}

	if (match_string (str, "?")) {
		if (evalf_or (str, &v1)) {
			return (1);
		}

		if (match_string (str, ":") == 0) {
			return (1);
		}

		if (evalf_or (str, &v2)) {
			return (1);
		}

		*val = (*val >= 0.0) ? v1 : v2;
	}

	return (0);
}

static
int evalf_expr (const char **str, double *val)
{
	if (evalf_if (str, val)) {
		return (1);
	}

	return (0);
}

static
int evalf_str (const char *str)
{
	double val;

	val = 0.0;

	if (evalf_expr (&str, &val)) {
		return (1);
	}

	printf ("%.20g\n", val);

	return (0);
}

int main (int argc, char *argv[])
{
	int i;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("evalf");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "e", "expression")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing expression\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (evalf_str (argv[i])) {
				fprintf (stderr, "%s: bad expression (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (argv[i][0] != '-') {
			if (evalf_str (argv[i])) {
				fprintf (stderr, "%s: bad expression (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv0, argv[i]
			);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	return (EXIT_SUCCESS);
}
