/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     fmtmu.c                                                    *
 * Created:       2004-08-03 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-04-22 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2004-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: fmtmu.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include "fmtmu.h"


typedef int (*write_line_f) (line_t *ln, FILE *out, int last);


static const char    *argv0 = NULL;

static int           par_keep = 0;
static unsigned      par_width = 78;
static unsigned      par_margin = 0;
static unsigned      par_max_space = 0;

static unsigned long par_line = 0;

static char          *par_stop = ".,!?";


static
void prt_help (void)
{
	fputs (
		"Usage: fmtmu [options] [input [output]]\n"
		"  -i string             Set input file name [stdin]\n"
		"  -o string             Set output file name [stdout]\n"
		"  -k, --keep            Keep format lines [no]\n"
		"  -m, --margin int      Set the left margin [0]\n"
		"  -w, --width int       Set the line width [78]\n"
		"  -s, --stop string     Set stop characters [.,!?]\n"
		"  -x, --max-spaces int  Set maximum number of spaces [0]\n",
		stdout
	);

	fflush (stdout);
}

static
void prt_char (FILE *fp, unsigned char c, unsigned n)
{
	while (n > 0) {
		fputc (c, fp);
		n -= 1;
	}
}

static
void line_init (line_t *ln)
{
	ln->i = 0;
	ln->n = 0;
	ln->buf = NULL;

	ln->eol_cnt = 0;
}

static
void line_free (line_t *ln)
{
	free (ln->buf);
}

static
int line_copy (line_t *dst, const line_t *src)
{
	unsigned char *tmp;

	if (dst->n < src->i) {
		tmp = realloc (dst->buf, src->i);
		if (tmp == NULL) {
			return (1);
		}
		dst->buf = tmp;
		dst->n = src->i;
	}

	dst->i = src->i;
	memcpy (dst->buf, src->buf, src->i);

	dst->eol_cnt = src->eol_cnt;
	memcpy (dst->eol, src->eol, src->eol_cnt);

	return (0);
}

static
int line_add_chars (line_t *ln, const unsigned char *buf, unsigned cnt)
{
	unsigned      i, j;
	unsigned char *tmp;

	j = ln->i + cnt;

	if (j > ln->n) {
		i = ln->i;
		while (i < j) {
			i = 2 * i + 1;
		}

		tmp = realloc (ln->buf, i);
		if (tmp == NULL) {
			return (1);
		}

		ln->buf = tmp;
		ln->n = i;
	}

	memcpy (ln->buf + ln->i, buf, cnt);
	ln->i += cnt;

	return (0);
}

static
int line_add_char (line_t *ln, unsigned char c)
{
	return (line_add_chars (ln, &c, 1));
}

static
int line_read (line_t *ln, FILE *fp)
{
	int c;

	par_line += 1;

	ln->i = 0;
	ln->eol_cnt = 0;

	c = fgetc (fp);
	while (c != EOF) {
		switch (c) {
			case 0x0d:
				ln->eol[0] = 0x0d;
				ln->eol_cnt = 1;

				c = fgetc (fp);
				if (c == 0x0a) {
					ln->eol[1] = 0x0a;
					ln->eol_cnt = 2;
				}
				else {
					ungetc (c, fp);
				}
				return (0);

			case 0x0a:
				ln->eol[0] = 0x0a;
				ln->eol_cnt = 1;
				return (0);

			default:
				if (line_add_char (ln, c)) {
					return (1);
				}
				break;
		}

		c = fgetc (fp);
	}

	return (0);
}

static
int line_write_eol (line_t *ln, FILE *fp)
{
	if (ln->eol_cnt > 0) {
		if (fwrite (ln->eol, 1, ln->eol_cnt, fp) != ln->eol_cnt) {
			return (1);
		}
	}

	return (0);
}

static
int line_write (line_t *ln, FILE *fp)
{
	if (ln->i > 0) {
		if (fwrite (ln->buf, 1, ln->i, fp) != ln->i) {
			return (1);
		}
	}

	if (ln->eol_cnt > 0) {
		if (fwrite (ln->eol, 1, ln->eol_cnt, fp) != ln->eol_cnt) {
			return (1);
		}
	}

	return (0);
}

static
int line_set_eol (line_t *ln, const unsigned char *buf, unsigned n)
{
	if (n > 16) {
		return (1);
	}

	memcpy (ln->eol, buf, n);
	ln->eol_cnt = n;

	return (0);
}

static
int char_is_space (unsigned char c)
{
	if ((c == ' ') || (c == '\t')) {
		return (1);
	}

	return (0);
}

static
int char_is_stop (unsigned char c)
{
	unsigned i;

	i = 0;
	while (par_stop[i] != 0) {
		if (c == par_stop[i]) {
			return (1);
		}

		i += 1;
	}

	return (0);
}

static
void line_compact (line_t *ln)
{
	int      space;
	unsigned i, j;

	i = 0;
	j = 0;
	space = 0;

	while (i < ln->i) {
		if (char_is_space (ln->buf[i])) {
			space = 1;
		}
		else {
			if (space && (j > 0) ) {
				ln->buf[j++] = ' ';
			}
			ln->buf[j++] = ln->buf[i];
			space = 0;
		}

		i += 1;
	}

	ln->i = j;
}

static
int line_is_eof (line_t *ln)
{
	if ((ln->i == 0) && (ln->eol_cnt == 0)) {
		return (1);
	}

	return (0);
}

static
int line_get_uint (line_t *ln, const char *id, unsigned *val)
{
	unsigned i;
	unsigned idn;

	idn = strlen (id);

	i = 0;
	while ((i + idn) <= ln->i) {
		if (strncmp ((char *) ln->buf + i, id, idn) == 0) {
			break;
		}

		i += 1;
	}

	i += idn;

	if (i >= ln->i) {
		return (1);
	}

	*val = 0;

	while ((i < ln->i) && char_is_space (ln->buf[i])) {
		i += 1;
	}

	while (i < ln->i) {
		if ((ln->buf[i] >= '0') && (ln->buf[i] <= '9')) {
			*val = 10 * *val + (ln->buf[i] - '0');
		}
		else {
			break;
		}

		i += 1;
	}

	return (0);
}


static
int write_line_c (line_t *ln, FILE *out, int last)
{
	unsigned n;

	n = (ln->i < par_width) ? ((par_width - ln->i) / 2) : 0;

	prt_char (out, ' ', par_margin + n);
	line_write (ln, out);

	return (0);
}

static
int write_line_j (line_t *ln, FILE *out, int last)
{
	unsigned i, n;
	unsigned sc, wc, pc;

	prt_char (out, ' ', par_margin);

	if (last) {
		line_write (ln, out);
		return (0);
	}

	wc = 0;
	pc = 0;

	for (i = 0; i < ln->i; i++) {
		if (ln->buf[i] == ' ') {
			wc += 1;

			if ((i > 0) && char_is_stop (ln->buf[i - 1])) {
				pc += 1;
			}
		}
	}

	sc = (ln->i < par_width) ? (par_width - ln->i) : 0;

	if (sc > pc) {
		if ((par_max_space > 0) && ((sc - pc) > par_max_space)) {
			sc = pc + par_max_space;
		}
	}

	for (i = 0; i < ln->i; i++) {
		fputc (ln->buf[i], out);

		n = 0;

		if (ln->buf[i] == ' ') {
			if (sc > pc) {
				n = (sc - pc) / wc;
				sc -= n;
			}

			if ((i > 0) && char_is_stop (ln->buf[i - 1])) {
				if (sc >= pc) {
					n += 1;
					sc -= 1;
				}

				pc -= 1;
			}

			wc -= 1;

			prt_char (out, ' ', n);
		}
	}

	if (ln->eol_cnt > 0) {
		if (fwrite (ln->eol, 1, ln->eol_cnt, out) != ln->eol_cnt) {
			return (1);
		}
	}

	return (0);
}

static
int write_line_l (line_t *ln, FILE *out, int last)
{
	prt_char (out, ' ', par_margin);
	line_write (ln, out);

	return (0);
}

static
int write_line_r (line_t *ln, FILE *out, int last)
{
	unsigned n;

	n = (ln->i < par_width) ? (par_width - ln->i) : 0;

	prt_char (out, ' ', par_margin + n);
	line_write (ln, out);

	return (0);
}

static
int format_block (FILE *inp, FILE *out, line_t *ln, write_line_f wl)
{
	int           c;
	unsigned char word[256];
	unsigned      n;
	unsigned      id0, id1, id2;
	unsigned      save_margin, save_width;

	if (line_get_uint (ln, "m1=", &id1)) {
		id1 = 0;
	}

	if (line_get_uint (ln, "m2=", &id2)) {
		id2 = 0;
	}

	if (line_get_uint (ln, "m0=", &id0)) {
		id0 = id1;
	}

	if ((id1 + id2) >= par_width) {
		return (1);
	}

	if ((id0 + id2) >= par_width) {
		return (1);
	}

	save_margin = par_margin;
	save_width = par_width;

	par_margin = save_margin + id0;
	par_width = save_width - id0 - id2;

	ln->i = 0;
	n = 0;

	c = fgetc (inp);
	while (c != EOF) {
		if (c == 0x0d) {
			c = fgetc (inp);
			if (c != 0x0a) {
				ungetc (c, inp);
				c = 0x0a;
			}
		}

		if (c == 0x0a) {
			c = fgetc (inp);
			ungetc (c, inp);
			if ((c == 0x0a) || (c == '%')) {
				break;
			}

			c = ' ';
		}

		if (char_is_space (c)) {
			if (n > 0) {
				if ((ln->i + n + 1) >= par_width) {
					wl (ln, out, 0);
					ln->i = 0;
					par_margin = save_margin + id1;
					par_width = save_width - id1 - id2;
				}
				if (ln->i > 0) {
					line_add_char (ln, ' ');
				}
				line_add_chars (ln, word, n);
				n = 0;
			}
		}
		else {
			word[n++] = c;
		}

		c = fgetc (inp);
	}

	if (n > 0) {
		if ((ln->i + n + 1) >= par_width) {
			wl (ln, out, 0);
			ln->i = 0;
			par_margin = save_margin + id1;
			par_width = save_width - id1 - id2;
		}
		if (ln->i > 0) {
			line_add_char (ln, ' ');
		}
		line_add_chars (ln, word, n);
	}

	wl (ln, out, 1);

	par_margin = save_margin;
	par_width = save_width;

	if (line_read (ln, inp)) {
		return (1);
	}

	return (0);
}

int format_bc (FILE *inp, FILE *out, line_t *ln)
{
	while (line_read (ln, inp) == 0) {
		if (line_is_eof (ln)) {
			return (0);
		}

		if ((ln->i >= 1) && (ln->buf[0] == '%')) {
			return (0);
		}

		line_compact (ln);

		if (ln->i == 0) {
			return (0);
		}

		write_line_c (ln, out, 0);
	}

	return (0);
}

int format_bj (FILE *inp, FILE *out, line_t *ln)
{
	int    first;
	line_t tmp;

	line_init (&tmp);

	first = 1;

	while (line_read (ln, inp) == 0) {
		if (line_is_eof (ln)) {
			break;
		}

		if ((ln->i >= 1) && (ln->buf[0] == '%')) {
			break;
		}

		line_compact (ln);

		if (ln->i == 0) {
			break;
		}

		if (first == 0) {
			write_line_j (&tmp, out, 0);
		}

		first = 0;
		line_copy (&tmp, ln);
	}

	if (first == 0) {
		prt_char (out, ' ', par_margin);
		line_write (&tmp, out);
	}

	line_free (&tmp);

	return (0);
}

int format_bl (FILE *inp, FILE *out, line_t *ln)
{
	while (line_read (ln, inp) == 0) {
		if (line_is_eof (ln)) {
			return (0);
		}

		if ((ln->i >= 1) && (ln->buf[0] == '%')) {
			return (0);
		}

		line_compact (ln);

		if (ln->i == 0) {
			return (0);
		}

		write_line_l (ln, out, 0);
	}

	return (0);
}

int format_br (FILE *inp, FILE *out, line_t *ln)
{
	while (line_read (ln, inp) == 0) {
		if (line_is_eof (ln)) {
			return (0);
		}

		if ((ln->i >= 1) && (ln->buf[0] == '%')) {
			return (0);
		}

		line_compact (ln);

		if (ln->i == 0) {
			return (0);
		}

		write_line_r (ln, out, 0);
	}

	return (0);
}

int format_b (FILE *inp, FILE *out, line_t *ln)
{
	if (ln->i >= 3) {
		switch (ln->buf[2]) {
			case 'c':
				return (format_bc (inp, out, ln));

			case 'j':
				return (format_bj (inp, out, ln));

			case 'l':
				return (format_bl (inp, out, ln));

			case 'r':
				return (format_br (inp, out, ln));
		}
	}

	fprintf (stderr, "line %lu: unknown sequence %%b%c\n",
		par_line, (ln->i >= 3) ? ln->buf[2] : ' '
	);

	return (1);
}

int format_f (FILE *inp, FILE *out, line_t *ln)
{
	if ((ln->i >= 3) && (ln->buf[2] == 'f')) {
		if (par_keep == 0) {
			fputs ("\x0c", out);
		}

		if (line_read (ln, inp)) {
			return (1);
		}

		while ((ln->i == 0) && (ln->eol_cnt > 0)) {
			if (par_keep) {
				line_write (ln, out);
			}

			if (line_read (ln, inp)) {
				return (1);
			}
		}

		return (0);
	}
	else if ((ln->i >= 3) && (ln->buf[2] == 'h')) {
		unsigned char c;
		unsigned      m1, m2;

		if (line_get_uint (ln, "m1=", &m1)) {
			m1 = 0;
		}

		if (line_get_uint (ln, "m2=", &m2)) {
			m2 = 0;
		}

		if (line_read (ln, inp)) {
			return (1);
		}

		line_compact (ln);

		c = (ln->i >= 1) ? ln->buf[0] : '-';

		prt_char (out, ' ', par_margin + m1);

		if ((m1 + m2) < par_width) {
			prt_char (out, c, par_width - m1 - m2);
		}
		else {
			prt_char (out, c, 1);
		}

		line_write_eol (ln, out);

		if (line_read (ln, inp)) {
			return (1);
		}

		return (0);
	}

	fprintf (stderr, "line %lu: unknown sequence %%f%c\n",
		par_line, (ln->i >= 3) ? ln->buf[2] : ' '
	);

	return (1);
}

int format_p (FILE *inp, FILE *out, line_t *ln)
{
	if (ln->i >= 3) {
		switch (ln->buf[2]) {
			case 'c':
				return (format_block (inp, out, ln, &write_line_c));

			case 'j':
				return (format_block (inp, out, ln, &write_line_j));

			case 'l':
				return (format_block (inp, out, ln, &write_line_l));

			case 'r':
				return (format_block (inp, out, ln, &write_line_r));
		}
	}

	fprintf (stderr, "line %lu: unknown sequence %%p%c\n",
		par_line, (ln->i >= 3) ? ln->buf[2] : ' '
	);

	return (1);
}

int format (FILE *inp, FILE *out)
{
	line_t ln;

	line_init (&ln);

	if (line_read (&ln, inp)) {
		return (1);
	}

	while (line_is_eof (&ln) == 0) {
		if ((ln.i >= 1) && (ln.buf[0] == '%')) {
			if (par_keep) {
				line_write (&ln, out);
			}

			if ((ln.i >= 2) && (ln.buf[1] == 'b')) {
				if (format_b (inp, out, &ln)) {
					return (1);
				}
			}
			else if ((ln.i >= 2) && (ln.buf[1] == 'f')) {
				if (format_f (inp, out, &ln)) {
					return (1);
				}
			}
			else if ((ln.i >= 2) && (ln.buf[1] == 'p')) {
				if (format_p (inp, out, &ln)) {
					return (1);
				}
			}
			else if (ln.i == 1) {
				if (line_read (&ln, inp)) {
					return (1);
				}
			}
			else {
				fprintf (stderr, "line %lu: unknown sequence %%%c\n",
					par_line, ln.buf[1]
				);

				return (1);
			}
		}
		else {
			if (ln.i > 0) {
				prt_char (out, ' ', par_margin);
			}

			line_write (&ln, out);

			if (line_read (&ln, inp)) {
				return (1);
			}
		}
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int      i, r;
	unsigned fcnt;
	char     *finp, *fout;
	FILE     *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("fmtmu");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "k", "keep")) {
			par_keep = 1;
		}
		else if (str_isarg2 (argv[i], "w", "width")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing width\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_width)) {
				fprintf (stderr, "%s: bad width (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "m", "margin")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing margin\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_margin)) {
				fprintf (stderr, "%s: bad margin (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "s", "--stop")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing stop characters\n", argv0);
				return (EXIT_FAILURE);
			}

			par_stop = argv[i];
		}
		else if (str_isarg2 (argv[i], "x", "max-spaces")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing maximum number of spaces\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_max_space)) {
				fprintf (stderr, "%s: bad maximum number of spaces (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_width < 4) {
		par_width = 4;
	}

	inp = (finp == NULL) ? stdin : fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
		return (EXIT_FAILURE);
	}

	out = (fout == NULL) ? stdout : fopen (fout, "wb");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
		return (EXIT_FAILURE);
	}

	r = format (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
