/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     hex.c                                                      *
 * Created:       2000-12-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2000-2007 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: hex.c 461 2007-10-15 21:25:26Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include "hex.h"


static const char    *argv0 = NULL;

static unsigned      par_dump = 1;
static int           par_dump_c = 0;
static unsigned      par_cols = 16;
static unsigned      par_base = 16;
static unsigned      par_digits = 2;
static int           par_chars = 1;
static int           par_lcase = 0;
static unsigned      par_spaces = 1;
static unsigned long par_addr = 0;


static
void prt_help (void)
{
	fputs (
		"hex: hex dump of binary files\n\n"
		"usage: hex [options] [input [output]]\n"
		"  -a, --address int  Set the base address [0]\n"
		"  -b, --base int     Set base for dump [16]\n"
		"  -c, --columns int  Set number of bytes per row [16]\n"
		"  -C, --dump-c       Use a C style dump format\n"
		"  -d, --decode       Reconstruct a binary file from a hex dump\n"
		"  -e, --encode       Dump a binary file [default]\n"
		"  -i string          Set input file name [stdin]\n"
		"  -l, --lower-case   Use lower case letters [no]\n"
		"  -n, --digits int   Set number of digits per byte [2]\n"
		"  -o string          Set output file name [stdout]\n"
		"  -p, --spaces int   Set number of spaces between bytes [1]\n"
		"  -s, --short        Select short output format [no]\n",
		stdout
	);

	fflush (stdout);
}

static
int is_digit (unsigned c)
{
	unsigned d;

	d = par_base;

	if ((c >= '0') && (c <= '9')) {
		d = (unsigned)(c - '0');
	}
	else if ((c >= 'a') && (c <= 'z')) {
		d = (unsigned)(10 + c - 'a');
	}
	else if ((c >= 'A') && (c <= 'Z')) {
		d = (unsigned)(10 + c - 'A');
	}

	return (d < par_base);
}

static
unsigned dig_to_int (unsigned char c)
{
	if ((c >= '0') && (c <= '9')) {
		return (c - '0');
	}
	else if ((c >= 'a') && (c <= 'z')) {
		return (c + 10 - 'a');
	}
	else if ((c >= 'A') && (c <= 'Z')) {
		return (c + 10 - 'A');
	}

	return (0);
}

static
void set_char (FILE *fp, unsigned c)
{
	unsigned i, d;
	char     buf[HEX_MAX_DIG];

	i = 0;

	do {
		d = c % par_base;
		c = c / par_base;

		if (par_lcase) {
			d += (d < 10) ? '0' : ('a' - 10);
		}
		else {
			d += (d < 10) ? '0' : ('A' - 10);
		}

		buf[i++] = (char) d;
		if (i >= HEX_MAX_DIG) {
			break;
		}
	} while ((c != 0) || (i < par_digits));

	while (i > 0) {
		fputc (buf[--i], fp);
	}
}

static
void set_char_c (FILE *fp, unsigned c)
{
	fprintf (fp, "0x%02x", c & 0xff);
}

static
void set_line (FILE *fp, unsigned long ofs, unsigned char *ln, unsigned cnt)
{
	unsigned i, j;

	if (cnt == 0) {
		return;
	}

	for (i = 1; i < cnt; i++) {
		set_char (fp, ln[i - 1]);
		for (j = 0; j < par_spaces; j++) {
			fputc (' ', fp);
		}
	}

	set_char (fp, ln[cnt - 1]);

	if (par_chars) {
		fprintf (fp, "\t# %08lX  ", par_addr + ofs);

		for (i = 0; i < cnt; i++) {
			if ((ln[i] >= 32) && (ln[i] < 127)) {
				fputc (ln[i], fp);
			}
			else {
				fputc ('.', fp);
			}
		}
	}

	fputc ('\n', fp);
}

static
void set_line_c (FILE *fp, unsigned long ofs, unsigned char *ln, unsigned cnt,
	int last)
{
	unsigned i;

	if (cnt == 0) {
		return;
	}

	for (i = 1; i < cnt; i++) {
		set_char_c (fp, ln[i - 1]);
		fputs (", ", fp);
	}

	set_char_c (fp, ln[cnt - 1]);

	if (last == 0) {
		fputs (",", fp);
	}

	if (par_chars) {
		fprintf (fp, "\t/* %08lX */", ofs);
	}

	fputc ('\n', fp);
}

static
int do_dump (FILE *inp, FILE *out)
{
	unsigned      col;
	unsigned long ofs;
	unsigned char buf[HEX_MAX_COL];
	int           next;

	ofs = 0;

	next = fgetc (inp);

	while (next != EOF) {
		buf[0] = (unsigned char) next;
		col = 1 + fread (buf + 1, 1, par_cols - 1, inp);
		next = fgetc (inp);

		if (par_dump_c) {
			set_line_c (out, ofs, buf, col, next == EOF);
		}
		else {
			set_line (out, ofs, buf, col);
		}

		ofs += col;
	}

	fflush (out);

	return (0);
}

static
int do_dump_stream (FILE *inp, FILE *out)
{
	int      c;
	unsigned i;

	c = fgetc (inp);
	if (c != EOF) {
		set_char (out, c);
	}

	c = fgetc (inp);
	while (c != EOF) {
		for (i = 0; i < par_spaces; i++) {
			fputc (' ', out);
		}

		set_char (out, c);

		c = fgetc (inp);
	}

	fflush (out);

	return (0);
}

static
int undump_digit (FILE *inp, FILE *out, int c)
{
	unsigned i, v;

	i = 0;
	v = 0;

	while (is_digit (c)) {
		v = (par_base * v + dig_to_int (c)) & 0xff;

		i += 1;
		if ((par_digits > 0) && (i >= par_digits)) {
			break;
		}

		c = fgetc (inp);
		if (c == EOF) {
			break;
		}
	}

	fputc (v, out);

	return (0);
}

static
int undump_string (FILE *inp, FILE *out)
{
	int c;

	while (1) {
		c = fgetc (inp);
		if (c == EOF) {
			return (1);
		}

		if (c == '"') {
			return (0);
		}

		fputc (c, out);
	}

	return (0);
}

static
int do_undump (FILE *inp, FILE *out)
{
	int c;

	while (1) {
		c = fgetc (inp);
		if (c == EOF) {
			break;
		}

		if (is_digit (c)) {
			if (undump_digit (inp, out, c)) {
				return (1);
			}
		}
		else if (c == '"') {
			if (undump_string (inp, out)) {
				return (1);
			}
		}
		else if (c == '#') {
			do {
				c = fgetc (inp);
			} while ((c != 0x0a) && (c != 0x0d) && (c != EOF));
		}
	}

	fflush (out);

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	int        have_digits = 0;
	unsigned   fcnt;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("hex");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "e", "encode")) {
			par_dump = 1;
		}
		else if (str_isarg2 (argv[i], "d", "decode")) {
			par_dump = 0;
			if (have_digits == 0) {
				par_digits = 0;
			}
		}
		else if (str_isarg2 (argv[i], "C", "dump-c")) {
			par_dump_c = 1;
		}
		else if (str_isarg2 (argv[i], "s", "short")) {
			par_chars = 0;
		}
		else if (str_isarg2 (argv[i], "a", "address")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing address\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_addr)) {
				fprintf (stderr, "%s: bad address (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "b", "base")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing base\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_base)) {
				fprintf (stderr, "%s: bad base (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "c", "columns")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing column count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_cols)) {
				fprintf (stderr, "%s: bad column count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "l", "lower-case")) {
			par_lcase = 1;
		}
		else if (str_isarg2 (argv[i], "n", "digits")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing digit count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_digits)) {
				fprintf (stderr, "%s: bad digit count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			have_digits = 1;
		}
		else if (str_isarg2 (argv[i], "p", "spaces")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing spaces count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_spaces)) {
				fprintf (stderr, "%s: bad spaces count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_cols > HEX_MAX_COL) {
		par_cols = HEX_MAX_COL;
	}

	if (par_base < 2) {
		par_base = 2;
	}
	else if (par_base > 36) {
		par_base = 36;
	}

	if (par_digits > HEX_MAX_DIG) {
		par_digits = HEX_MAX_DIG;
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, par_dump ? "rb" : "r");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, par_dump ? "w" : "wb");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	if (par_dump) {
		if (par_cols == 0) {
			r = do_dump_stream (inp, out);
		}
		else {
			r = do_dump (inp, out);
		}
	}
	else {
		r = do_undump (inp, out);
	}

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
