/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     rmd160.c                                                   *
 * Created:       2005-02-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-02-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005 Hampa Hug <hampa@hampa.ch>                        *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: rmd160.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/config.h>
#include <lib/rmd160.h>


#define rmd_f1(x, y, z) ((x) ^ (y) ^ (z))
#define rmd_f2(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define rmd_f3(x, y, z) (((x) | ~(y)) ^ (z))
#define rmd_f4(x, y, z) (((x) & (z)) | ((y) & ~(z)))
#define rmd_f5(x, y, z) ((x) ^ ((y) | ~(z)))

#define rmd_rol(x, n) (((x) << (n)) | ((x) >> (32 - (n))))

#define rmd_r1(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f1 ((b), (c), (d)) + (M)) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_r2(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f2 ((b), (c), (d)) + (M) + 0x5a827999) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_r3(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f3 ((b), (c), (d)) + (M) + 0x6ed9eba1) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_r4(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f4 ((b), (c), (d)) + (M) + 0x8f1bbcdc) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_r5(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f5 ((b), (c), (d)) + (M) + 0xa953fd4e) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_p1(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f1 ((b), (c), (d)) + (M)) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_p2(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f2 ((b), (c), (d)) + (M) + 0x7a6d76e9) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_p3(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f3 ((b), (c), (d)) + (M) + 0x6d703ef3) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_p4(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f4 ((b), (c), (d)) + (M) + 0x5c4dd124) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)

#define rmd_p5(a, b, c, d, e, M, s) do { \
		(a) = ((a) + rmd_f5 ((b), (c), (d)) + (M) + 0x50a28be6) & 0xffffffff; \
		(a) = (rmd_rol ((a), (s)) + (e)) & 0xffffffff; \
		(c) = rmd_rol ((c), 10) & 0xffffffff; \
	} while (0)


void rmd160_init (rmd160_t *rmd)
{
	rmd->buf_i = 0;

	rmd->cnt[0] = 0;
	rmd->cnt[1] = 0;

	rmd->h[0] = 0x67452301UL;
	rmd->h[1] = 0xefcdab89UL;
	rmd->h[2] = 0x98badcfeUL;
	rmd->h[3] = 0x10325476UL;
	rmd->h[4] = 0xc3d2e1f0UL;
}

void rmd160_copy (rmd160_t *dst, const rmd160_t *src)
{
	unsigned i;

	dst->buf_i = src->buf_i;

	for (i = 0; i < src->buf_i; i++) {
		dst->buf[i] = src->buf[i];
	}

	dst->cnt[0] = src->cnt[0];
	dst->cnt[1] = src->cnt[1];

	for (i = 0; i < 5; i++) {
		dst->h[i] = src->h[i];
	}
}

static
void rmd_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = v & 0xff;
	tmp[1] = (v >> 8) & 0xff;
	tmp[2] = (v >> 16) & 0xff;
	tmp[3] = (v >> 24) & 0xff;
}

static
void rmd160_add_size (rmd160_t *rmd, unsigned long n)
{
	n &= 0xffffffffUL;

	rmd->cnt[0] = (rmd->cnt[0] + n) & 0xffffffffUL;

	if (rmd->cnt[0] < n) {
		rmd->cnt[1] = (rmd->cnt[1] + 1) & 0xffffffffUL;
	}
}

static
void rmd_blk_frm_buf (unsigned long *dst, const unsigned char *src)
{
	unsigned      i;
	unsigned long v;

	for (i = 0; i < 16; i++) {
		v = src[3];
		v = (v << 8) + src[2];
		v = (v << 8) + src[1];
		v = (v << 8) + src[0];

		dst[i] = v;

		src += 4;
	}
}

static
void rmd160_calc_block (rmd160_t *rmd, const void *buf)
{
	unsigned long a1, b1, c1, d1, e1;
	unsigned long a2, b2, c2, d2, e2;
	unsigned long t;
	unsigned long M[16];

	rmd_blk_frm_buf (M, buf);

	a1 = rmd->h[0];
	b1 = rmd->h[1];
	c1 = rmd->h[2];
	d1 = rmd->h[3];
	e1 = rmd->h[4];

	a2 = rmd->h[0];
	b2 = rmd->h[1];
	c2 = rmd->h[2];
	d2 = rmd->h[3];
	e2 = rmd->h[4];

	rmd_r1 (a1, b1, c1, d1, e1, M[ 0], 11);
	rmd_r1 (e1, a1, b1, c1, d1, M[ 1], 14);
	rmd_r1 (d1, e1, a1, b1, c1, M[ 2], 15);
	rmd_r1 (c1, d1, e1, a1, b1, M[ 3], 12);
	rmd_r1 (b1, c1, d1, e1, a1, M[ 4],  5);
	rmd_r1 (a1, b1, c1, d1, e1, M[ 5],  8);
	rmd_r1 (e1, a1, b1, c1, d1, M[ 6],  7);
	rmd_r1 (d1, e1, a1, b1, c1, M[ 7],  9);
	rmd_r1 (c1, d1, e1, a1, b1, M[ 8], 11);
	rmd_r1 (b1, c1, d1, e1, a1, M[ 9], 13);
	rmd_r1 (a1, b1, c1, d1, e1, M[10], 14);
	rmd_r1 (e1, a1, b1, c1, d1, M[11], 15);
	rmd_r1 (d1, e1, a1, b1, c1, M[12],  6);
	rmd_r1 (c1, d1, e1, a1, b1, M[13],  7);
	rmd_r1 (b1, c1, d1, e1, a1, M[14],  9);
	rmd_r1 (a1, b1, c1, d1, e1, M[15],  8);

	rmd_r2 (e1, a1, b1, c1, d1, M[ 7],  7);
	rmd_r2 (d1, e1, a1, b1, c1, M[ 4],  6);
	rmd_r2 (c1, d1, e1, a1, b1, M[13],  8);
	rmd_r2 (b1, c1, d1, e1, a1, M[ 1], 13);
	rmd_r2 (a1, b1, c1, d1, e1, M[10], 11);
	rmd_r2 (e1, a1, b1, c1, d1, M[ 6],  9);
	rmd_r2 (d1, e1, a1, b1, c1, M[15],  7);
	rmd_r2 (c1, d1, e1, a1, b1, M[ 3], 15);
	rmd_r2 (b1, c1, d1, e1, a1, M[12],  7);
	rmd_r2 (a1, b1, c1, d1, e1, M[ 0], 12);
	rmd_r2 (e1, a1, b1, c1, d1, M[ 9], 15);
	rmd_r2 (d1, e1, a1, b1, c1, M[ 5],  9);
	rmd_r2 (c1, d1, e1, a1, b1, M[ 2], 11);
	rmd_r2 (b1, c1, d1, e1, a1, M[14],  7);
	rmd_r2 (a1, b1, c1, d1, e1, M[11], 13);
	rmd_r2 (e1, a1, b1, c1, d1, M[ 8], 12);

	rmd_r3 (d1, e1, a1, b1, c1, M[ 3], 11);
	rmd_r3 (c1, d1, e1, a1, b1, M[10], 13);
	rmd_r3 (b1, c1, d1, e1, a1, M[14],  6);
	rmd_r3 (a1, b1, c1, d1, e1, M[ 4],  7);
	rmd_r3 (e1, a1, b1, c1, d1, M[ 9], 14);
	rmd_r3 (d1, e1, a1, b1, c1, M[15],  9);
	rmd_r3 (c1, d1, e1, a1, b1, M[ 8], 13);
	rmd_r3 (b1, c1, d1, e1, a1, M[ 1], 15);
	rmd_r3 (a1, b1, c1, d1, e1, M[ 2], 14);
	rmd_r3 (e1, a1, b1, c1, d1, M[ 7],  8);
	rmd_r3 (d1, e1, a1, b1, c1, M[ 0], 13);
	rmd_r3 (c1, d1, e1, a1, b1, M[ 6],  6);
	rmd_r3 (b1, c1, d1, e1, a1, M[13],  5);
	rmd_r3 (a1, b1, c1, d1, e1, M[11], 12);
	rmd_r3 (e1, a1, b1, c1, d1, M[ 5],  7);
	rmd_r3 (d1, e1, a1, b1, c1, M[12],  5);

	rmd_r4 (c1, d1, e1, a1, b1, M[ 1], 11);
	rmd_r4 (b1, c1, d1, e1, a1, M[ 9], 12);
	rmd_r4 (a1, b1, c1, d1, e1, M[11], 14);
	rmd_r4 (e1, a1, b1, c1, d1, M[10], 15);
	rmd_r4 (d1, e1, a1, b1, c1, M[ 0], 14);
	rmd_r4 (c1, d1, e1, a1, b1, M[ 8], 15);
	rmd_r4 (b1, c1, d1, e1, a1, M[12],  9);
	rmd_r4 (a1, b1, c1, d1, e1, M[ 4],  8);
	rmd_r4 (e1, a1, b1, c1, d1, M[13],  9);
	rmd_r4 (d1, e1, a1, b1, c1, M[ 3], 14);
	rmd_r4 (c1, d1, e1, a1, b1, M[ 7],  5);
	rmd_r4 (b1, c1, d1, e1, a1, M[15],  6);
	rmd_r4 (a1, b1, c1, d1, e1, M[14],  8);
	rmd_r4 (e1, a1, b1, c1, d1, M[ 5],  6);
	rmd_r4 (d1, e1, a1, b1, c1, M[ 6],  5);
	rmd_r4 (c1, d1, e1, a1, b1, M[ 2], 12);

	rmd_r5 (b1, c1, d1, e1, a1, M[ 4],  9);
	rmd_r5 (a1, b1, c1, d1, e1, M[ 0], 15);
	rmd_r5 (e1, a1, b1, c1, d1, M[ 5],  5);
	rmd_r5 (d1, e1, a1, b1, c1, M[ 9], 11);
	rmd_r5 (c1, d1, e1, a1, b1, M[ 7],  6);
	rmd_r5 (b1, c1, d1, e1, a1, M[12],  8);
	rmd_r5 (a1, b1, c1, d1, e1, M[ 2], 13);
	rmd_r5 (e1, a1, b1, c1, d1, M[10], 12);
	rmd_r5 (d1, e1, a1, b1, c1, M[14],  5);
	rmd_r5 (c1, d1, e1, a1, b1, M[ 1], 12);
	rmd_r5 (b1, c1, d1, e1, a1, M[ 3], 13);
	rmd_r5 (a1, b1, c1, d1, e1, M[ 8], 14);
	rmd_r5 (e1, a1, b1, c1, d1, M[11], 11);
	rmd_r5 (d1, e1, a1, b1, c1, M[ 6],  8);
	rmd_r5 (c1, d1, e1, a1, b1, M[15],  5);
	rmd_r5 (b1, c1, d1, e1, a1, M[13],  6);

	rmd_p5 (a2, b2, c2, d2, e2, M[ 5],  8);
	rmd_p5 (e2, a2, b2, c2, d2, M[14],  9);
	rmd_p5 (d2, e2, a2, b2, c2, M[ 7],  9);
	rmd_p5 (c2, d2, e2, a2, b2, M[ 0], 11);
	rmd_p5 (b2, c2, d2, e2, a2, M[ 9], 13);
	rmd_p5 (a2, b2, c2, d2, e2, M[ 2], 15);
	rmd_p5 (e2, a2, b2, c2, d2, M[11], 15);
	rmd_p5 (d2, e2, a2, b2, c2, M[ 4],  5);
	rmd_p5 (c2, d2, e2, a2, b2, M[13],  7);
	rmd_p5 (b2, c2, d2, e2, a2, M[ 6],  7);
	rmd_p5 (a2, b2, c2, d2, e2, M[15],  8);
	rmd_p5 (e2, a2, b2, c2, d2, M[ 8], 11);
	rmd_p5 (d2, e2, a2, b2, c2, M[ 1], 14);
	rmd_p5 (c2, d2, e2, a2, b2, M[10], 14);
	rmd_p5 (b2, c2, d2, e2, a2, M[ 3], 12);
	rmd_p5 (a2, b2, c2, d2, e2, M[12],  6);

	rmd_p4 (e2, a2, b2, c2, d2, M[ 6],  9);
	rmd_p4 (d2, e2, a2, b2, c2, M[11], 13);
	rmd_p4 (c2, d2, e2, a2, b2, M[ 3], 15);
	rmd_p4 (b2, c2, d2, e2, a2, M[ 7],  7);
	rmd_p4 (a2, b2, c2, d2, e2, M[ 0], 12);
	rmd_p4 (e2, a2, b2, c2, d2, M[13],  8);
	rmd_p4 (d2, e2, a2, b2, c2, M[ 5],  9);
	rmd_p4 (c2, d2, e2, a2, b2, M[10], 11);
	rmd_p4 (b2, c2, d2, e2, a2, M[14],  7);
	rmd_p4 (a2, b2, c2, d2, e2, M[15],  7);
	rmd_p4 (e2, a2, b2, c2, d2, M[ 8], 12);
	rmd_p4 (d2, e2, a2, b2, c2, M[12],  7);
	rmd_p4 (c2, d2, e2, a2, b2, M[ 4],  6);
	rmd_p4 (b2, c2, d2, e2, a2, M[ 9], 15);
	rmd_p4 (a2, b2, c2, d2, e2, M[ 1], 13);
	rmd_p4 (e2, a2, b2, c2, d2, M[ 2], 11);

	rmd_p3 (d2, e2, a2, b2, c2, M[15],  9);
	rmd_p3 (c2, d2, e2, a2, b2, M[ 5],  7);
	rmd_p3 (b2, c2, d2, e2, a2, M[ 1], 15);
	rmd_p3 (a2, b2, c2, d2, e2, M[ 3], 11);
	rmd_p3 (e2, a2, b2, c2, d2, M[ 7],  8);
	rmd_p3 (d2, e2, a2, b2, c2, M[14],  6);
	rmd_p3 (c2, d2, e2, a2, b2, M[ 6],  6);
	rmd_p3 (b2, c2, d2, e2, a2, M[ 9], 14);
	rmd_p3 (a2, b2, c2, d2, e2, M[11], 12);
	rmd_p3 (e2, a2, b2, c2, d2, M[ 8], 13);
	rmd_p3 (d2, e2, a2, b2, c2, M[12],  5);
	rmd_p3 (c2, d2, e2, a2, b2, M[ 2], 14);
	rmd_p3 (b2, c2, d2, e2, a2, M[10], 13);
	rmd_p3 (a2, b2, c2, d2, e2, M[ 0], 13);
	rmd_p3 (e2, a2, b2, c2, d2, M[ 4],  7);
	rmd_p3 (d2, e2, a2, b2, c2, M[13],  5);

	rmd_p2 (c2, d2, e2, a2, b2, M[ 8], 15);
	rmd_p2 (b2, c2, d2, e2, a2, M[ 6],  5);
	rmd_p2 (a2, b2, c2, d2, e2, M[ 4],  8);
	rmd_p2 (e2, a2, b2, c2, d2, M[ 1], 11);
	rmd_p2 (d2, e2, a2, b2, c2, M[ 3], 14);
	rmd_p2 (c2, d2, e2, a2, b2, M[11], 14);
	rmd_p2 (b2, c2, d2, e2, a2, M[15],  6);
	rmd_p2 (a2, b2, c2, d2, e2, M[ 0], 14);
	rmd_p2 (e2, a2, b2, c2, d2, M[ 5],  6);
	rmd_p2 (d2, e2, a2, b2, c2, M[12],  9);
	rmd_p2 (c2, d2, e2, a2, b2, M[ 2], 12);
	rmd_p2 (b2, c2, d2, e2, a2, M[13],  9);
	rmd_p2 (a2, b2, c2, d2, e2, M[ 9], 12);
	rmd_p2 (e2, a2, b2, c2, d2, M[ 7],  5);
	rmd_p2 (d2, e2, a2, b2, c2, M[10], 15);
	rmd_p2 (c2, d2, e2, a2, b2, M[14],  8);

	rmd_p1 (b2, c2, d2, e2, a2, M[12],  8);
	rmd_p1 (a2, b2, c2, d2, e2, M[15],  5);
	rmd_p1 (e2, a2, b2, c2, d2, M[10], 12);
	rmd_p1 (d2, e2, a2, b2, c2, M[ 4],  9);
	rmd_p1 (c2, d2, e2, a2, b2, M[ 1], 12);
	rmd_p1 (b2, c2, d2, e2, a2, M[ 5],  5);
	rmd_p1 (a2, b2, c2, d2, e2, M[ 8], 14);
	rmd_p1 (e2, a2, b2, c2, d2, M[ 7],  6);
	rmd_p1 (d2, e2, a2, b2, c2, M[ 6],  8);
	rmd_p1 (c2, d2, e2, a2, b2, M[ 2], 13);
	rmd_p1 (b2, c2, d2, e2, a2, M[13],  6);
	rmd_p1 (a2, b2, c2, d2, e2, M[14],  5);
	rmd_p1 (e2, a2, b2, c2, d2, M[ 0], 15);
	rmd_p1 (d2, e2, a2, b2, c2, M[ 3], 13);
	rmd_p1 (c2, d2, e2, a2, b2, M[ 9], 11);
	rmd_p1 (b2, c2, d2, e2, a2, M[11], 11);

	t = rmd->h[1] + c1 + d2;
	rmd->h[1] = (rmd->h[2] + d1 + e2) & 0xffffffff;
	rmd->h[2] = (rmd->h[3] + e1 + a2) & 0xffffffff;
	rmd->h[3] = (rmd->h[4] + a1 + b2) & 0xffffffff;
	rmd->h[4] = (rmd->h[0] + b1 + c2) & 0xffffffff;
	rmd->h[0] = t & 0xffffffff;
}

void rmd160_calc (rmd160_t *rmd, const void *buf, unsigned long cnt)
{
	unsigned            i, n;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	if (rmd->buf_i > 0) {
		while ((rmd->buf_i < 64) && (cnt > 0)) {
			rmd->buf[rmd->buf_i++] = *(tmp++);
			cnt -= 1;
		}

		if (rmd->buf_i < 64) {
			return;
		}

		rmd160_calc_block (rmd, rmd->buf);

		rmd->buf_i = 0;
		rmd160_add_size (rmd, 8 * 64);
	}

	while (cnt >= 64) {
		rmd160_calc_block (rmd, tmp);

		rmd160_add_size (rmd, 8 * 64);

		tmp += 64;
		cnt -= 64;
	}

	n = (unsigned) cnt;
	for (i = 0; i < n; i++) {
		rmd->buf[i] = tmp[i];
	}

	rmd->buf_i = n;
}

void rmd160_done (rmd160_t *rmd)
{
	unsigned i;

	rmd160_add_size (rmd, 8 * rmd->buf_i);

	rmd->buf[rmd->buf_i++] = 0x80;

	if (rmd->buf_i > 56) {
		for (i = rmd->buf_i; i < 64; i++) {
			rmd->buf[i] = 0;
		}

		rmd160_calc_block (rmd, rmd->buf);

		rmd->buf_i = 0;
	}

	for (i = rmd->buf_i; i < 56; i++) {
		rmd->buf[i] = 0;
	}

	rmd_set_uint32 (rmd->buf, 56, rmd->cnt[0]);
	rmd_set_uint32 (rmd->buf, 60, rmd->cnt[1]);

	rmd160_calc_block (rmd, rmd->buf);
}

unsigned long rmd160_get_sum_i (rmd160_t *rmd, unsigned i)
{
	if (i < 5) {
		return (rmd->h[i]);
	}

	return (0);
}

void rmd160_get_sum (rmd160_t *rmd, void *buf)
{
	rmd_set_uint32 (buf, 0, rmd->h[0]);
	rmd_set_uint32 (buf, 4, rmd->h[1]);
	rmd_set_uint32 (buf, 8, rmd->h[2]);
	rmd_set_uint32 (buf, 12, rmd->h[3]);
	rmd_set_uint32 (buf, 16, rmd->h[4]);
}

void rmd160_calc1 (void *dst, const void *src, unsigned long n)
{
	rmd160_t rmd;

	rmd160_init (&rmd);
	rmd160_calc (&rmd, src, n);
	rmd160_done (&rmd);
	rmd160_get_sum (&rmd, dst);
}
