/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     sha256.c                                                   *
 * Created:       2005-07-27 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-07-27 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005 Hampa Hug <hampa@hampa.ch>                        *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: sha256.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/config.h>
#include <lib/sha256.h>


static unsigned long sha256_K[64] = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
	0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
	0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
	0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
	0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
	0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
	0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
	0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
	0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};


static inline
unsigned long sha256_add (unsigned long x, unsigned long y)
{
	return ((x + y) & 0xffffffff);
}

static inline
unsigned long sha256_ror (unsigned long x, unsigned n)
{
	x = (x >> n) | (x << (32 - n));

	return (x & 0xffffffff);
}

static inline
unsigned long sha256_S0 (unsigned long x)
{
	return (sha256_ror (x, 2) ^ sha256_ror (x, 13) ^ sha256_ror (x, 22));
}

static inline
unsigned long sha256_S1 (unsigned long x)
{
	return (sha256_ror (x, 6) ^ sha256_ror (x, 11) ^ sha256_ror (x, 25));
}

static inline
unsigned long sha256_s0 (unsigned long x)
{
	return (sha256_ror (x, 7) ^ sha256_ror (x, 18) ^ (x >> 3));
}

static inline
unsigned long sha256_s1 (unsigned long x)
{
	return (sha256_ror (x, 17) ^ sha256_ror (x, 19) ^ (x >> 10));
}

static inline
unsigned long sha256_F (unsigned long x, unsigned long y, unsigned long z)
{
	return (((x & y) ^ (~x & z)) & 0xffffffff);
}

static inline
unsigned long sha256_G (unsigned long x, unsigned long y, unsigned long z)
{
	return (((x & y) ^ (x & z) ^ (y & z)) & 0xffffffff);
}


void sha256_init (sha256_t *sha)
{
	sha->buf_i = 0;

	sha->cnt[0] = 0;
	sha->cnt[1] = 0;

	sha->H[0] = 0x6a09e667;
	sha->H[1] = 0xbb67ae85;
	sha->H[2] = 0x3c6ef372;
	sha->H[3] = 0xa54ff53a;
	sha->H[4] = 0x510e527f;
	sha->H[5] = 0x9b05688c;
	sha->H[6] = 0x1f83d9ab;
	sha->H[7] = 0x5be0cd19;
}

void sha256_copy (sha256_t *dst, const sha256_t *src)
{
	unsigned i;

	dst->buf_i = src->buf_i;

	for (i = 0; i < src->buf_i; i++) {
		dst->buf[i] = src->buf[i];
	}

	dst->cnt[0] = src->cnt[0];
	dst->cnt[1] = src->cnt[1];

	for (i = 0; i < 8; i++) {
		dst->H[i] = src->H[i];
	}
}

static inline
void sha256_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 24) & 0xff;
	tmp[1] = (v >> 16) & 0xff;
	tmp[2] = (v >> 8) & 0xff;
	tmp[3] = v & 0xff;
}

static inline
unsigned long sha256_get_uint32 (const void *buf, unsigned i)
{
	unsigned long       v;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	v = tmp[0];
	v = (v << 8) | tmp[1];
	v = (v << 8) | tmp[2];
	v = (v << 8) | tmp[3];

	return (v);
}

static
void sha256_add_size (sha256_t *sha, unsigned long n)
{
	n &= 0xffffffffUL;

	sha->cnt[0] = (sha->cnt[0] + n) & 0xffffffff;

	if (sha->cnt[0] < n) {
		sha->cnt[1] = (sha->cnt[1] + 1) & 0xffffffff;
	}
}

static
void sha256_blk_frm_buf (unsigned long *W, const void *buf)
{
	unsigned i;

	for (i = 0; i < 16; i++) {
		W[i] = sha256_get_uint32 (buf, 4 * i);
	}

	for (i = 16; i < 64; i++) {
		W[i] = sha256_s1 (W[i - 2]) + W[i - 7] + sha256_s0 (W[i - 15]) + W[i - 16];
		W[i] &= 0xffffffff;
	}
}

static
void sha256_calc_block (sha256_t *sha, const void *buf)
{
	unsigned      i;
	unsigned long h[8];
	unsigned long W[64];
	unsigned long t1, t2;

	sha256_blk_frm_buf (W, buf);

	for (i = 0; i < 8; i++) {
		h[i] = sha->H[i];
	}

	for (i = 0; i < 64; i++) {
		t1 = h[7] + sha256_S1 (h[4]) + sha256_F (h[4], h[5], h[6]) + sha256_K[i] + W[i];
		t2 = sha256_S0 (h[0]) + sha256_G (h[0], h[1], h[2]);

		h[7] = h[6];
		h[6] = h[5];
		h[5] = h[4];
		h[4] = sha256_add (h[3], t1);
		h[3] = h[2];
		h[2] = h[1];
		h[1] = h[0];
		h[0] = sha256_add (t1, t2);
	}

	for (i = 0; i < 8; i++) {
		sha->H[i] = sha256_add (sha->H[i], h[i]);
	}
}

void sha256_calc (sha256_t *sha, const void *buf, unsigned long cnt)
{
	unsigned            i, n;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	if (sha->buf_i > 0) {
		while ((sha->buf_i < 64) && (cnt > 0)) {
			sha->buf[sha->buf_i] = *tmp;
			tmp += 1;
			sha->buf_i += 1;
			cnt -= 1;
		}

		if (sha->buf_i < 64) {
			return;
		}

		sha256_calc_block (sha, sha->buf);

		sha->buf_i = 0;
		sha256_add_size (sha, 8 * 64);
	}

	while (cnt >= 64) {
		sha256_calc_block (sha, tmp);
		sha256_add_size (sha, 8 * 64);

		tmp += 64;
		cnt -= 64;
	}

	n = (unsigned) cnt;
	for (i = 0; i < n; i++) {
		sha->buf[i] = tmp[i];
	}

	sha->buf_i = n;
}

void sha256_done (sha256_t *sha)
{
	unsigned i;

	sha256_add_size (sha, 8 * sha->buf_i);

	sha->buf[sha->buf_i++] = 0x80;

	if (sha->buf_i > 56) {
		for (i = sha->buf_i; i < 64; i++) {
			sha->buf[i] = 0;
		}

		sha256_calc_block (sha, sha->buf);

		sha->buf_i = 0;
	}

	for (i = sha->buf_i; i < 56; i++) {
		sha->buf[i] = 0;
	}

	sha256_set_uint32 (sha->buf, 56, sha->cnt[1]);
	sha256_set_uint32 (sha->buf, 60, sha->cnt[0]);

	sha256_calc_block (sha, sha->buf);

	sha->buf_i = 0;
}

unsigned long sha256_get_sum_i (sha256_t *sha, unsigned i)
{
	if (i < 8) {
		return (sha->H[i]);
	}

	return (0);
}

int sha256_get_sum (sha256_t *sha, void *buf, unsigned max)
{
	unsigned i;

	if (max < 32) {
		return (1);
	}

	for (i = 0; i < 8; i++) {
		sha256_set_uint32 (buf, 4 * i, sha->H[i]);
	}

	return (0);
}

void sha256_calc1 (void *dst, const void *src, unsigned long n)
{
	sha256_t sha;

	sha256_init (&sha);
	sha256_calc (&sha, src, n);
	sha256_done (&sha);
	sha256_get_sum (&sha, dst, 32);
}
