/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     sched.c                                                    *
 * Created:       2002-03-21 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-07-11 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: sched.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <sched.h>

#include "sched.h"


static const char *argv0 = NULL;

static int   par_verbose = 0;
static char  *par_cmd = NULL;
static char  **par_arg = NULL;

static int   par_policy = SCHED_OTHER;
static char  *par_policy_name = "SCHED_OTHER";
static int   par_priority = 0;


static
void prt_help (void)
{
	fputs (
		"Usage: sched [options]\n"
		"  --help                Print usage information\n"
		"  --version             Print version information\n"
		"  -v, --verbose         Verbose operation\n"
		"  -q, --query           Print out current scheduling policy and exit\n"
		"  -p, --policy string   Set scheduling policy\n"
		"  -n, --priority int    Set static priority\n"
		"  -c, --command string  Set command to execute\n"
		"Policies are:\n"
		"  other\n"
		"  fifo\n"
		"  rr\n",
		stdout
	);

	fflush (stdout);
}

static
int do_query (void)
{
	int                pol;
	int                p_max, p_min;
	char               *str;
	struct sched_param p;

	pol = sched_getscheduler (0);

	if (sched_getparam (0, &p)) {
		fprintf (stderr, "%s: could not get schduling parameters\n", argv0);
		return (1);
	}

	p_min = sched_get_priority_min (pol);
	p_max = sched_get_priority_max (pol);

	switch (pol) {
		case SCHED_OTHER:
			str = "SCHED_OTHER";
			break;

		case SCHED_FIFO:
			str = "SCHED_FIFO";
			break;

		case SCHED_RR:
			str = "SCHED_RR";
			break;

		default:
			str = "unknown";
			break;
	};

	fprintf (stdout,
		"policy: %s (%d)  priority: %d  min: %d  max: %d\n",
		str, pol, p.sched_priority, p_min, p_max
	);

	return (0);
}

static
int do_sched (void)
{
	struct sched_param p;

	if (sched_getparam (0, &p)) {
		fprintf (stderr, "%s: could not get schduling parameters\n", argv0);
		return (1);
	}

	p.sched_priority = par_priority;

	if (sched_setscheduler (0, par_policy, &p)) {
		fprintf (stderr, "%s: can't set scheduling policy\n", argv0);
		return (1);
	}

	execvp (par_cmd, par_arg);

	fprintf (stderr, "%s: can't exec command (%s)\n", argv0, par_cmd);

	return (1);
}

int main (int argc, char *argv[])
{
	int i, r;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("sched");
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], "q", "query")) {
			r = do_query();
			if (r) {
				return (EXIT_FAILURE);
			}
			else {
				return (EXIT_SUCCESS);
			}
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else if (str_isarg2 (argv[i], "p", "policy")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing policy\n", argv0);
				return (EXIT_FAILURE);
			}

			if (strcmp (argv[i], "other") == 0) {
				par_policy = SCHED_OTHER;
				par_policy_name = "SCHED_OTHER";
			}
			else if (strcmp (argv[i], "fifo") == 0) {
				par_policy = SCHED_FIFO;
				par_policy_name = "SCHED_FIFO";
			}
			else if (strcmp (argv[i], "rr") == 0) {
				par_policy = SCHED_RR;
				par_policy_name = "SCHED_RR";
			}
			else {
				if (str_get_sint (argv[i], &par_policy)) {
					fprintf (stderr, "%s: bad scheduling policy (%s)\n", argv0, argv[i]);
					return (EXIT_FAILURE);
				}

				par_policy_name = "numeric";
			}
		}
		else if (str_isarg2 (argv[i], "n", "priority")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing priority\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_isarg1 (argv[i], "min")) {
				par_priority = sched_get_priority_min (par_policy);
			}
			else if (str_isarg1 (argv[i], "max")) {
				par_priority = sched_get_priority_max (par_policy);
			}
			else {
				if (str_get_sint (argv[i], &par_priority)) {
					fprintf (stderr, "%s: bad priority (%s)\n", argv0, argv[i]);
					return (EXIT_FAILURE);
				}
			}
		}
		else if (str_isarg2 (argv[i], "c", "command")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing command\n", argv0);
				return (EXIT_FAILURE);
			}

			par_cmd = argv[i];
			par_arg = &argv[i];

			i = argc;
		}
		else {
			fprintf (stderr, "%s: unknown parameter (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_verbose) {
		fprintf (stderr, "policy:   %s (%d)\n", par_policy_name, par_policy);
		fprintf (stderr, "priority: %d\n", par_priority);
	}

	if (par_cmd == NULL) {
		fprintf (stderr, "%s: need a command\n", argv0);
		return (1);
	}

	r = do_sched();

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
